/*
 * generated by Xtext
 */
package com.framsticks.framclipse.scoping;

import static com.framsticks.framclipse.framScript.FramScriptPackage.Literals.ARRAY_ELEMENT_EXPRESSION__ARRAY;
import static com.framsticks.framclipse.framScript.FramScriptPackage.Literals.QUALIFIED_EXPRESSION__CHILD;
import static com.framsticks.framclipse.framScript.FramScriptPackage.Literals.QUALIFIED_EXPRESSION__PARENT;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceImpl;
import org.eclipse.xtext.resource.EObjectDescription;
import org.eclipse.xtext.resource.IEObjectDescription;
import org.eclipse.xtext.resource.IResourceDescription;
import org.eclipse.xtext.scoping.IGlobalScopeProvider;
import org.eclipse.xtext.scoping.IScope;
import org.eclipse.xtext.scoping.Scopes;
import org.eclipse.xtext.scoping.impl.AbstractDeclarativeScopeProvider;
import org.eclipse.xtext.scoping.impl.SimpleScope;

import com.framsticks.framclipse.framScript.Block;
import com.framsticks.framclipse.framScript.Code;
import com.framsticks.framclipse.framScript.Expdef;
import com.framsticks.framclipse.framScript.ForEachStatement;
import com.framsticks.framclipse.framScript.ForStatement;
import com.framsticks.framclipse.framScript.FramScriptFactory;
import com.framsticks.framclipse.framScript.Function;
import com.framsticks.framclipse.framScript.GotoStatment;
import com.framsticks.framclipse.framScript.Header;
import com.framsticks.framclipse.framScript.IncludeDeclaration;
import com.framsticks.framclipse.framScript.Invocation;
import com.framsticks.framclipse.framScript.LabeledStatement;
import com.framsticks.framclipse.framScript.Model;
import com.framsticks.framclipse.framScript.PropertyIncludeDeclaration;
import com.framsticks.framclipse.framScript.ProposableFunctionImpl;
import com.framsticks.framclipse.framScript.ProposableVariableDeclarationImpl;
import com.framsticks.framclipse.framScript.QualifiedExpression;
import com.framsticks.framclipse.framScript.Show;
import com.framsticks.framclipse.framScript.Statement;
import com.framsticks.framclipse.framScript.VariableDeclaration;
import com.framsticks.framclipse.framScript.VariableDeclarations;
import com.framsticks.framclipse.framScript.VariableRef;
import com.framsticks.framclipse.resource.FramScriptResourceDescriptionManager;
import com.framsticks.framclipse.script.ConstantProvider;
import com.framsticks.framclipse.script.ExpressionTraverser;
import com.framsticks.framclipse.script.model.Element;
import com.framsticks.framclipse.script.model.Type;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.inject.Inject;

/**
 * This class contains custom scoping description.
 * 
 * See https://www.eclipse.org/Xtext/documentation/303_runtime_concepts.html#scoping
 * on how and when to use it.
 */
public class FramScriptScopeProvider extends AbstractDeclarativeScopeProvider {

	private final ConstantProvider provider;

	private final ExpressionTraverser traverser;

	@Inject
	private IGlobalScopeProvider globalScopeProvider;

	@Inject
	private FramScriptResourceDescriptionManager descriptionManager;

	private Resource resource;

	private Map<String, IScope> typesScopes = Maps.newHashMap();
	private Map<Type, IScope> functionScopes = Maps.newHashMap();
	private Map<Type, IScope> fieldScopes = Maps.newHashMap();

	private FramScriptImportUriResolver resolver = new FramScriptImportUriResolver();

	@Inject
	public FramScriptScopeProvider(ConstantProvider provider, ExpressionTraverser traverser) {
		this.provider = provider;
		this.traverser = traverser;
		this.resource = new ResourceImpl(URI.createURI("/"));
		initTypes();
	}

	@Override
	public IScope getScope(EObject context, EReference reference) {
		if (reference.getEContainingClass().getName().equals("VariableRef") &&
				reference.getName().equals("var")) {
			if (context instanceof VariableRef) {
				QualifiedExpression parent = getParent((VariableRef) context);
				if (parent != null) {
					return getScope(parent, reference);
				}
				EObject object = context;
				while (object != null && !(object.eContainer() instanceof Block ||
						object.eContainer() instanceof ForStatement ||
						object.eContainer() instanceof ForEachStatement)) {
					object = object.eContainer();
				}
				Statement statement = (Statement) object;
				EObject blockOrFor = object.eContainer();

				if (blockOrFor instanceof Block) {
					Block block = (Block) blockOrFor;
					return Scopes.scopeFor(precedingDeclarations(block, statement),
							nextValidScope(block, reference));
				} else
					return getScope(blockOrFor, reference);
			} else if (context.eContainer() instanceof Block) {
				Block block = (Block) context.eContainer();
				List<VariableDeclaration> vars = precedingDeclarations(block, (Statement)context);
				IScope outerScope = Scopes.scopeFor(vars, nextValidScope(block, reference));
				if (context instanceof Block) {
					return outerScope;
				}
				return new SimpleScope(outerScope,
						super.getScope(context, reference).getAllElements());
			}
		}
		return super.getScope(context, reference);
	}

	private IScope nextValidScope(EObject context, EReference ref) {
		if (ref.getEContainingClass().getName().equals("VariableRef") &&
				ref.getName().equals("var")) {
			if (context instanceof Block) {
				EObject parent = context.eContainer();
				if (parent == null) {	// may happen in .inc files.
					return scope_VariableRef_var((Model)context, ref);
				}
				if (!(parent instanceof Block)) {
					return getScope(parent, ref);
				}
			}
		}
		return getScope(context, ref);
	}

	private List<VariableDeclaration> precedingDeclarations(Block block, Statement statement){
		List<VariableDeclaration> vars = Lists.newArrayList();
		for (Statement stmt : block.getStatements()) {
			// Everything after the statement is out of scope.
			if (stmt == statement) {
				break;
			} else if (stmt instanceof VariableDeclarations) {
				vars.addAll(((VariableDeclarations) stmt).getVars());
			}
		}
		return vars;
	}

	protected void initTypes() {
		List<Function> allFunctions = Lists.newArrayList();
		List<VariableDeclaration> allFields = Lists.newArrayList();
		
		for (Entry<String, Map<String, Type>> entry : provider.getTypes().entrySet()) {
			List<VariableDeclaration> types = Lists.newArrayList();	
			for (Type type : entry.getValue().values()) {
				types.add(createType(type));
				
				ArrayList<Function> functions = Lists.newArrayList();
				ArrayList<VariableDeclaration> fields = Lists.newArrayList();
				for (Element element : type.getElements()) {
					if (element.isFunction()) {
						functions.add(createFunction(element));
					} else {
						fields.add(createField(element));
					}
				}
				functionScopes.put(type, Scopes.scopeFor(functions));
				fieldScopes.put(type, Scopes.scopeFor(fields));
				allFunctions.addAll(functions);
				allFields.addAll(fields);
			}
			typesScopes.put(entry.getKey(), Scopes.scopeFor(types));
		}
		
		functionScopes.put(null, Scopes.scopeFor(allFunctions));
		fieldScopes.put(null, Scopes.scopeFor(allFields));
	}
	
	protected IScope scope_Invocation_function(Invocation inv, EReference ref) {
		QualifiedExpression parent = getParent(inv);
		IScope scope;
		if (parent != null) {
			scope = scope_Invocation_function(parent, ref);
		} else {
			scope = scope_Invocation_function(getFunction(inv), ref);
		}
		Show show = getShow(inv);
		IScope showScope = show == null ? IScope.NULLSCOPE :
			new SimpleScope(filterByArgumentsCount(getScope(show, ref), inv.getArgs().size()));
		return new SimpleScope(showScope, filterByArgumentsCount(scope, inv.getArgs().size()));
	}

	protected IScope scope_Invocation_function(QualifiedExpression inv, EReference ref) {
		Type type = traverser.getLastType(inv);
		IScope scope = functionScopes.get(type);
		return scope != null ? scope : IScope.NULLSCOPE;
	}

	protected IScope scope_Invocation_function(Function f, EReference ref) {
		IScope scope = delegateGetScope(f, ref);
		scope = addAliases(scope);
		return scope;
	}

	protected IScope scope_Invocation_function(Show show, EReference ref) {
		return addAliases(new SimpleScope(showExpdefIncludes(show, ref)));
	}

	/*
	 * Make sure scope contains only function descriptions.
	 */
	private List<IEObjectDescription> filterByArgumentsCount(IScope scope, int size) {
		List<IEObjectDescription> filtered = Lists.newArrayList();
		for (IEObjectDescription desc : scope.getAllElements()) {
			Function function = (Function) desc.getEObjectOrProxy();
			if (function.getParams().size() == size) {
				filtered.add(desc);
			}
		}
		return filtered;
	}

	private IScope addAliases(IScope scope) {
		Iterable<IEObjectDescription> contents = scope.getAllElements();
		List<IEObjectDescription> withAliases = Lists.newArrayList(contents);
		for (IEObjectDescription desc : contents) {
			Function function = (Function) desc.getEObjectOrProxy();
			for (String string : function.getAliases()) {
				withAliases.add(EObjectDescription.create(string, function));
			}
		}
		return new SimpleScope(withAliases);
	}

	protected IScope scope_VariableRef_var(QualifiedExpression var, EReference ref) {
		Type type = traverser.getLastType(var);
		IScope scope = fieldScopes.get(type);
		return new SimpleScope(nextValidScope(var.eContainer(), ref),
				scope != null ? scope.getAllElements() : new ArrayList<IEObjectDescription>());
	}

	protected IScope scope_VariableRef_var(Block block, EReference ref) {
		// Scope for Blocks is constructed earlier in getScope, so we stop here.
		return IScope.NULLSCOPE;
	}

	protected IScope scope_VariableRef_var(ForStatement fs, EReference ref) {
		List<VariableDeclaration> vars = Lists.newArrayList();
		Statement init = fs.getInit();
		if (init instanceof VariableDeclarations) {
			vars.addAll(((VariableDeclarations)init).getVars());
		}
		return Scopes.scopeFor(vars, getScope(fs.eContainer(), ref));
	}

	protected IScope scope_VariableRef_var(ForEachStatement fes, EReference ref) {
		List<VariableDeclaration> vars = Lists.newArrayList();
		if (fes.getDecl() == null)
			return getScope(fes.eContainer(), ref);
		else {
			vars.add(fes.getDecl());
			return Scopes.scopeFor(vars, getScope(fes.eContainer(), ref));
		}
	}

	/*
	 * Blocks are handled specially in getScope,
	 * and scope_VariableRef_var(Block, EReference) returns IScope.NULLSCOPE,
	 * which breaks the chain of scopes,
	 * so we have to ensure each unspecified scope_VariableRef_var call
	 * is redirected to getScope in this class, instead of super.getScope.
	 */
	protected IScope scope_VariableRef_var(EObject ctx, EReference ref) {
		EObject parent = ctx.eContainer();
		if (parent == null)
			return IScope.NULLSCOPE;
		else
			return getScope(parent, ref);
	}

	protected IScope scope_VariableRef_var(Function f, EReference ref) {
		List<VariableDeclaration> vars = Lists.newArrayList();
		vars.addAll(f.getParams());
		return Scopes.scopeFor(vars, getScope(f.eContainer(), ref));
	}

	protected IScope scope_VariableRef_var(Code code, EReference ref) {
		List<VariableDeclaration> vars = Lists.newArrayList();
		for (VariableDeclarations declarations : code.getGlobals()) {
			vars.addAll(declarations.getVars());
		}
		return Scopes.scopeFor(vars, getScope(code.eContainer(), ref));
	}

	private List<IEObjectDescription> showExpdefIncludes(Show show, EReference ref) {
		ArrayList<IEObjectDescription> result = Lists.newArrayList();
		// Get expdef header
		Header expdefHeader= null;
		for (Header h : show.getHeaders()) {
			if ("expdef:".equals(h.getName())) {
					expdefHeader = h;
					break;
			}
		}
		if (expdefHeader == null)
			return result;
		// Find resource by name
		ResourceSet resourceSet = show.eResource().getResourceSet();
		String expdefFile = resolver.resolve(expdefHeader);
		Resource expdefResource = findResourceByName(resourceSet, expdefFile);
		if (expdefResource == null)
			return result;
		// Mine include directives
		ArrayList<String> included = Lists.newArrayList();
		Expdef expdef = (Expdef) expdefResource.getContents().get(0);
		for (PropertyIncludeDeclaration pid : expdef.getPropertyImports())
			included.add(pid.getImportURI());
		for (IncludeDeclaration id : expdef.getCode().getIncludes())
			included.add(id.getImportURI());
		// Get globals
		for (String filename : included) {
			Resource includedResource = findResourceByName(resourceSet, filename);
			if (includedResource == null) {
				String showURI = show.eResource().getURI().toString();
				URI includeURI = URI.createURI(showURI.subSequence(
						0, showURI.lastIndexOf("/")+1) + filename);
				includedResource = resourceSet.getResource(
						includeURI, true);
				if (includedResource == null)
					continue;
			}
			IResourceDescription resourceDescription =
					descriptionManager.getResourceDescription(includedResource);
			result.addAll(Lists.newArrayList(resourceDescription.getExportedObjectsByType(ref.getEReferenceType())));
		}
		return result;
	}

	private Resource findResourceByName(ResourceSet resourceSet,
			String filename) {
		for (Resource res : resourceSet.getResources()) {
			if (res.getURI().toString().endsWith("/" + filename)) {
				return res;
			}
		}
		return null;
	}

	protected IScope scope_VariableRef_var(Model model, EReference ref) {
		IScope outerScope = globalScopeProvider.getScope(model.eResource(), ref, null);
		if (model instanceof Show) {
			outerScope = new SimpleScope(outerScope, showExpdefIncludes((Show)model, ref));
		}
		String ext = getExtension(model);
		if ("inc".equals(ext))	// There is no description for .inc in XML,
			ext = "script";	// so we pretend it is a .script.
		IScope types = typesScopes.get(ext);
		return new SimpleScope(outerScope, types.getAllElements());
	}

	protected IScope scope_GotoStatment_dest(GotoStatment stmt, EReference ref) {
		Block code = getFunction(stmt).getCode();
		List<LabeledStatement> labels = Lists.newArrayList();
		for (TreeIterator<EObject> i = code.eAllContents(); i.hasNext();) {
			EObject object = i.next();
			if (object instanceof LabeledStatement) {
				labels.add((LabeledStatement) object);
			}
		}
		return Scopes.scopeFor(labels);
	}

	protected IScope scope_Property(Show show, EReference ref) {
		return new SimpleScope(delegateGetScope(show, ref), showExpdefIncludes(show, ref));
	}

	protected IScope scope_State(Show show, EReference ref) {
		return new SimpleScope(delegateGetScope(show, ref), showExpdefIncludes(show, ref));
	}

	private Function createFunction(Element element) {
		Function function = new ProposableFunctionImpl(element);
		EList<VariableDeclaration> params = function.getParams();
		for (int i = 0; i < element.getArguments().size(); i++) {
			params.add(createVariableDeclaration(element.getArguments().get(i).getName(i)));
		}
		resource.getContents().add(function);
		return function;
	}

	private VariableDeclaration createField(Element element) {
		VariableDeclaration field = new ProposableVariableDeclarationImpl(element);
		resource.getContents().add(field);
		return field;
	}

	private VariableDeclaration createType(Type type) {
		VariableDeclaration global = new ProposableVariableDeclarationImpl(type);
		resource.getContents().add(global);
		return global;
	}

	private VariableDeclaration createVariableDeclaration(String name) {
		VariableDeclaration var = FramScriptFactory.eINSTANCE.createVariableDeclaration();
		var.setName(name);
		resource.getContents().add(var);
		return var;
	}

	private QualifiedExpression getParent(QualifiedExpression expr) {
		EStructuralFeature feature = expr.eContainingFeature();
		if (feature.equals(QUALIFIED_EXPRESSION__CHILD)) {
			return (QualifiedExpression) expr.eContainer();
		} else if (feature.equals(QUALIFIED_EXPRESSION__PARENT)
				|| feature.equals(ARRAY_ELEMENT_EXPRESSION__ARRAY)) {
			return getParent((QualifiedExpression) expr.eContainer());
		} else {
			return null;
		}
	}

	private Function getFunction(EObject object) {
		while (object != null && !(object instanceof Function)) {
			object = object.eContainer();
		}
		return (Function) object;
	}

	private Show getShow(EObject object) {
		while (object != null && !(object instanceof Show)) {
			object = object.eContainer();
		}
		return (Show) object;
	}

	private String getExtension(EObject object) {
		return object.eResource().getURI().fileExtension();
	}

}
