package com.framsticks.framclipse.script.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import com.thoughtworks.xstream.annotations.XStreamAlias;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;

public class Element {

	@XStreamAsAttribute
	private String name, type;

	@XStreamAsAttribute
	private boolean function;

	@XStreamAsAttribute
	private Double min, max;

	@XStreamAlias("default")
	@XStreamAsAttribute
	private String def;

	private String description;

	private List<Argument> arguments = new ArrayList<Argument>(0);

	public String getName() {
		return name;
	}

	public String getType() {
		return type;
	}

	public boolean isFunction() {
		return function;
	}

	public Double getMin() {
		return min;
	}

	public Double getMax() {
		return max;
	}

	public String getDef() {
		return def;
	}

	public String getDescription() {
		return description;
	}

	public List<Argument> getArguments() {
		return arguments != null ? arguments : Collections.<Argument> emptyList();
	}

	public String toString() {
		return name + " : " + type;
	};

	public String argumentsString(boolean includeType) {
		StringBuilder builder = new StringBuilder("(");
		if (!getArguments().isEmpty()) {
			for (int i = 0; i < getArguments().size(); i++) {
				Argument a = arguments.get(i);
				if (includeType) {
					builder.append(a.getType() + " ");
				}
				builder.append(a.getName(i) + ", ");
			}
			builder.delete(builder.length() - 2, builder.length());
		}
		return builder.append(")").toString();
	}

	public String returnType() {
		return type != null ? type : "?";
	}

}
